part of http_router;

class HttpRouter extends StreamConsumer<HttpRequest, bool> {

  Dictionary       _routes;
  HttpRouteHandler _fallback;

  HttpRouter(Map routes, this._fallback) {
    var root = new Dictionary.fromMap(routes);
    this._routes = root.fold(new Dictionary(), (memo, v, k) {
      var tmp = new Dictionary.fromMap(v);
      memo[this._keyToRegExp(k)] = tmp.fold(new Dictionary(), (memo, v, k) {
        var tmp = new Dictionary.fromMap(v);
        memo[k] = tmp.fold(new Dictionary(), (memo, v, k) {
          memo[this._keyToRegExp(k)] = v;
          return memo;
        });
        return memo;
      });
      return memo;
    });
  }

  Future<bool> addStream(Stream<HttpRequest> requests) {
    var completer = new Completer<bool>();
    requests.listen((HttpRequest request) {

      var route = this._getHostHeader(request).flatMap((host) {
        return this._getHostMethods(host);
      }).flatMap((methods) {
        return this._getMethod(request, methods);
      }).flatMap((routes) {
        return this._getRoute(request, routes);
      });

      this._handleRoute(request, route);
    });
    return completer.future;
  }

  Option<String> _getHostHeader(HttpRequest request) {
    var hosts = request.headers["host"];
    if (hosts == null) {
      return new None();
    } else {
      return new Some(hosts[0]);
    }
  }

  Option<Dictionary> _getHostMethods(String host) {
    return this._routes.findValue((value, key) {
      return key.hasMatch(host);
    });
  }

  Option<Dictionary> _getMethod(HttpRequest request, Dictionary methods) {
    var method = request.method;
    return methods.get(method);
  }

  Option<HttpRouteHandler> _getRoute(HttpRequest request, Dictionary routes) {
    var path = request.uri.path;
    return routes.findValue((value, key) {
      return key.hasMatch(path);
    });
  }

  void _handleRoute(HttpRequest request, Option<HttpRouteHandler> handler) {
    if (handler.nonEmpty()) {
      handler.get().handler(request);
    } else {
      this._fallback.handler(request);
    }
  }

  RegExp _keyToRegExp(String key) {
    if (key.startsWith('r:')) {
      return new RegExp(key.substring(2));
    } else {
      return new RegExp('^' + key + '\$');
    }
  }

  Future<bool> close() {
    var completer = new Completer<bool>();
    return completer.future;
  }

  Future<bool> consume(Stream<HttpRequest> requests) {
    return this.addStream(requests);
  }

}